
/// Tag builder functions.
library tags.builder;

import 'dart:html';

/// A function to build an html element.
///
/// An optional build function may be provided to build child elements.
typedef E Tag<E extends Element>([Build build, args]);

/// A function to build children of a [Tag].
///
/// Each tag function that is called from within a build function will append
/// its element as a child of the current [host].  The order of such calls is 
/// equal to the order in which the children are appended.
typedef void Build();

/// A function to construct an instance of a node.
///
/// All tag functions that are called from within a create function will _not_
/// append any elements to the node being constructed.  The [host] will be 
/// _null_ within the scope of this function.  This allows the side-effect free
/// usage of [Tag]s in class initializers.
typedef Node Create();

// Stack of nodes used to build trees.  
// The top of the stack is always the host for the current build scope.
final _scopes = new List<Node>();

// Whether or not the builder is currently in a [Build] function scope.
bool _building = false;

/// [Create] a node and [Build] its children.
///
/// The node constructed by [Create] will be the [host] when the [Build] 
/// function [b] is called.  All [Tag] functions that are called within the 
/// scope of [b] will append their element as a child of the [host].  [Tag] 
/// functions called within the scope of [c] are side-effect free.
Node build(Create c, Build b) {
  final bool outerBuilding = _building;
  var node;
  try {
    _building = false;
    node = c();
    assert(node != null);
  } finally {  
    _building = outerBuilding;  
  }
  if (outerBuilding && !_scopes.isEmpty) {
    _scopes.last.append(node);  
  }
  if (b == null) return node;
  _scopes.add(node);  
  try {
    _building = true;
    b();
  } finally {
    _building = outerBuilding;
    _scopes.removeLast();
  }
  return node;
}

/// The node currently being built; may be `null` if none.
Node get host {
  if (!_building || _scopes.isEmpty) return null;
  else return _scopes.last;
}
